  /*
  Для WeMos, связанной с Исполнительным Устройством.
  Приём от Ардуино режима работы. 
  Выдача сигналов на излучатель.
  Обработка входного сигнала полностью и передача результатов на Ардуино.
  */

  #include <ledc.h>                    // Подключение библиотеки ШИМ
  #include <esp_now.h>
  #include <Arduino.h>                 // библиотека Arduino 
  #include <WiFi.h>
  #include <esp_wifi.h>

                    // Новый MAC Address этой платы
  uint8_t newMACAddress[] = {0x30, 0x83, 0x98, 0xB2, 0x6E, 0xAF};    
 
 	 // МАС-АДРЕС платы, на которую отправляем данные. C8:C9:A3:6A:D9:64
  uint8_t broadcastAddress[] = {0xC8, 0xC9, 0xA3, 0x6A, 0xD9, 0x64};

  #define  STR_Pin          16         // назначение вывода STR
  #define  FO_Pin           18         // назначение вывода FO
  #define  DC_Pin           33         // назначение вывода DC
  #define  echoPinA         1          // назначение вывода echoA L4 (3500 mV)
  #define  echoPinB         2          // назначение вывода echoB L3 (2500 mV)
  #define  echoPinC         4	         // назначение вывода echoC L2 (1000 mV)
  #define  echoPinD         6          // назначение вывода echoD L1 (200 mV)

  #define  A0R_Pin         39          // назначение вывода A0R
  #define  A1R_Pin         37          // назначение вывода A1R

  #define  analog1Pin       8          // назначение вывода analog1Pin Uпит излучателя

  #define  MOSI_PIN        11          // назначение вывода MOSI_PIN
  #define  MISO_PIN         9          // назначение вывода MISO_PIN
  #define  SCK_PIN          7          // назначение вывода SCK_PIN
  #define  SS_PIN          12          // назначение вывода SS_PIN
  #define  CSB_PIN          5          // назначение вывода CSB_PIN

  #define  PUSK_Pin        35          // назначение вывода PUSK
  #define  START_Pin        3          // назначение вывода START


			       // Вводим переменные для строба ШИМ
  volatile int pwmfrequency = 20000;   // 20 кГц - 4; 10 кГц - 8; 5 кГц - 16 импульсов.
  volatile int target_duty = 256;      // Целевая скважность, до неё доходит.
  uint32_t scale = 1;                  // Количество импульсов, на которое изменяется скважность.
  uint32_t cycle_num = 1;              // Количество периодов, через которое будет достигнута целевая скважность.
  uint32_t dutyCycle = 255;            // Начальная скважность.
  uint32_t idle_level = 0;             // Уровень выходного сигнала при выключении фадинга.
  uint32_t freqCLK = 20000;            // Частота ШИМ

			 // Вводим переменные для частоты ШИМ
  const int ledPin0 = FO_Pin;          // GPIO18	FO_Pin
  const int ledChannel0 = 0;           // Канал 0
  const int resolution0 = 2;           // Разрешение 2 бит
  const int dutyCycle0 = 2;            // Скважность 2 
  const int dutyCycle1 = 0;            // Скважность 0
  const uint32_t freq0 = 80000;        // Частота выходного сигнала


  int IN_PIN;                          // Пин входа текущего канала

  float analog1 = 0.0;                 // объявление переменной analog1
  uint32_t start;                      // Время старта
  uint32_t now;                        // Текущее машинное время в мкс
  uint32_t prevUs;                     // Предыдущее время изменения состояния пина

  int count = 0;                       // Счётчик циклов измерения
  bool first,prevS,flag = 0;           // Первое, предыдущее состояние пина, флаг не первого измерения, флаг ошибки
  bool flagrec = 0;                    // флаг приёма данных

  #define LOG_SIZE 12                  // размер лога (LOG_SIZE ячеек)
  uint16_t us[LOG_SIZE];               // массив измерений. Для длительности отрезков между изменениями входа

			 // Вводим переменные для хранения отправляемых данных
  byte regkn2 = 0xFF;                  // регистр чтения кнопок управления каналами
  byte regm1 = 0xC6;                   // регистр хранения режима управления исполнительным устройством
  byte regm2 = 0x3C;                   // регистр храннения режима управления каналами
  byte reg_IN = 0xFD;                  // регистр пина индикации
  byte regout = 0x3A;                  // регистр выдачи на исполнительное устройство
  float voltage1 = 0.0;                // объявление переменной voltage1 
  uint16_t duration;                   // объявление переменной длительности сигнала от приёмника
  uint16_t durus[LOG_SIZE];            // массив измерений. Для текущего времени в момент изменения
  uint16_t arrey1[LOG_SIZE];           // массив измерений. Для передачи

 			 // Вводим переменные для хранения принимаемых данных
  byte  incomingReg1;
  byte  incomingReg2;
  byte  incomingReg3;
  byte  incomingReg4;
  byte  incomingReg5;
  float incomingVoltbat;
  uint16_t incomingDursig;
  uint16_t incomingArrey1[LOG_SIZE];
 
			// Переменная для хранения состояния отправки
  String success;
 
			// структуры для отправки
 			// (Должна совпадать со структурой на плате-приемнике)
  typedef struct struct_message {
  byte reg1;
  byte reg2;
  byte reg3;
  byte reg4;
  byte reg5;
  float voltbat;                       // напряжение батареи
  uint16_t dursig;                     // длительность сигнала от приёмника
  uint16_t arrey1[LOG_SIZE];           // массив измерений. 
  } struct_message;
 
  			// Создаем переменную для хранения отправляемого сообщения
  struct_message ISPReadings;
 
  			// То же, но для принимаемого сообщения
  struct_message incomingReadings;

  			// Callback-функция для получения состояния отправки

  esp_now_peer_info_t peerInfo;

  void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  //Serial.print("\r\nLast Packet Send Status:\t");
  //Serial.println(status == ESP_NOW_SEND_SUCCESS ? "Delivery Success" : "Delivery Fail");
  if (status ==0){
    success = "Delivery Success :)";
  }
  else{
    success = "Delivery Fail :(";
  }
  }
  		// То же, для индикации состояния приема данных
  void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) {
  memcpy(&incomingReadings, incomingData, sizeof(incomingReadings));
  //Serial.print("Bytes received: ");
  //Serial.println(len);
 
  				// приём данных
  regkn2 = incomingReadings.reg1;
  regm1 = incomingReadings.reg2;
  regm2 = incomingReadings.reg3;
  reg_IN = incomingReadings.reg4;
  regout = incomingReadings.reg5;
  voltage1 = incomingReadings.voltbat;
  duration = incomingReadings.dursig;
  for (int i = 0; i < LOG_SIZE; i++) {
  durus[i] = incomingReadings.arrey1[i];
  } 
  printIncomingReadings();
  initial();
  flagrec = 1;         		// устанавливаем в 1 флаг приёма данных 
  }

  void printIncomingReadings(){
          // Отображаем показания в мониторе порта
  Serial.println("INCOMING READINGS"); 
  Serial.print("regkn2: ");
  Serial.println(regkn2, BIN);
  Serial.print("regm1: ");
  Serial.println(regm1, BIN);
  Serial.print("regm2: ");
  Serial.println(regm2, BIN);
  Serial.print("reg_IN: ");
  Serial.println(reg_IN, BIN);
  Serial.print("regout: ");
  Serial.println(regout, BIN);
  }

  void printSendingData(){
  Serial.println("SENDING_DATA");
  Serial.print("reg_IN: ");
  Serial.println(reg_IN, BIN);
  Serial.print("count: ");
  Serial.println(count);
  Serial.print("voltage1: ");
  Serial.println(voltage1);
  Serial.print("duration: ");
  Serial.println(duration);
  Serial.println("durus: ");
  for (int i = 0; i < LOG_SIZE; i++) {
  Serial.println(durus[i]); 
  } }

  void setup() {

  pinMode(STR_Pin, OUTPUT);		      // назначение вывода STR_Pin выходом 
  pinMode(FO_Pin, OUTPUT);	        // назначение вывода FO_Pin выходом 
  pinMode(DC_Pin, OUTPUT);	        // назначение вывода DC_Pin выходом 
  pinMode(A0R_Pin, OUTPUT);	        // назначение вывода A0R_Pin выходом 
  pinMode(A1R_Pin, OUTPUT);	        // назначение вывода A1R_Pin выходом 

  pinMode(echoPinA, INPUT);		      // назначение вывода echoA входом
  pinMode(echoPinB, INPUT);         // назначение вывода echoB входом
  pinMode(echoPinC, INPUT);	        // назначение вывода echoC входом
  pinMode(echoPinD, INPUT);	        // назначение вывода echoD входом 
  pinMode(analog1Pin, INPUT);       // назначение вывода analog1Pin входом 

  pinMode(MOSI_PIN, OUTPUT);        // назначение вывода MOSI_PIN выходом
  pinMode(MISO_PIN, INPUT);	        // назначение вывода MISO_PIN входом
  pinMode(SCK_PIN, OUTPUT);	        // назначение вывода SCK_PIN выходом 
  pinMode(SS_PIN, OUTPUT); 	        // назначение вывода SS_PIN выходом 
  pinMode(CSB_PIN, OUTPUT);		      // назначение вывода CSB_PIN выходом 

  pinMode(PUSK_Pin, INPUT_PULLUP);	// назначение вывода PUSK_Pin входом
  pinMode(START_Pin, INPUT);        // назначение вывода START_Pin входом


  			// Конфигурация ШИМ
  ledc_timer_config_t ledc_timer;  
  ledc_timer.speed_mode   = LEDC_LOW_SPEED_MODE;
  ledc_timer.timer_num    = LEDC_TIMER_1;
  ledc_timer.duty_resolution = LEDC_TIMER_8_BIT;
  ledc_timer.freq_hz      = pwmfrequency;
  ledc_channel_config_t ledc_channel2;
  ledc_channel2.channel    = LEDC_CHANNEL_2;
  ledc_channel2.gpio_num   = STR_Pin;	// GPIO16 STR_Pin
  ledc_channel2.speed_mode = LEDC_LOW_SPEED_MODE;
  ledc_channel2.timer_sel  = LEDC_TIMER_1;
  ledc_channel2.duty       = dutyCycle;
  ledc_timer_config(&ledc_timer);
  ledc_channel_config(&ledc_channel2);
  ledc_fade_func_install(ESP_INTR_FLAG_SHARED);

       // конфигурируем функционирование LED PWM   
  ledcSetup(ledChannel0, freq0, resolution0); // Задаём номер канала, частоту, скважность

     // attach the channel to the GPIO to be controlled
  ledcAttachPin(ledPin0, ledChannel0);        // Привязываем номер вывода к номеру канала

 
  			// Запускаем монитор порта
  Serial.begin(115200);
  Serial.println();
  
         		// Выставляем режим работы Wi-Fi
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  WiFi.setTxPower(WIFI_POWER_MINUS_1dBm);      // Установка минимального уровня мощности WiFi

                // Устанавливаем новый MAC Address этой платы 
  esp_wifi_set_mac(WIFI_IF_STA, &newMACAddress[0]);

			// Инициализируем протокол ESP-NOW

  if (esp_now_init() != ESP_OK) {
  //Serial.println("Error initializing ESP-NOW");
  return;
  }
  			// Регистрируем callback-функцию для получения статуса отправки
  esp_now_register_send_cb(OnDataSent);
 
  			// Регистрируем пиры
  memcpy(peerInfo.peer_addr, broadcastAddress, 6);
  peerInfo.channel = 0;  
  peerInfo.encrypt = false;

  			// Добавляем пиры

  if (esp_now_add_peer(&peerInfo) != ESP_OK){
    //Serial.println("Failed to add peer");
    return;
  }
 
		// Регистрируем callback-функцию для получения статуса приема
  esp_now_register_recv_cb(OnDataRecv);
  }
 
  void loop() {

  if (flagrec == 1) {             // если пришла посылка от WeMosArd
  
  			    // Чтение данных от приёмника
  			// читаем время и текущий пин из IN_PIN

  now = micros();                 // Читаем текущее машинное время
  if (count < LOG_SIZE) {         // Если буфер не переполнен
  bool s = digitalRead(IN_PIN);		// Читаем текущее состояние пина
  if (prevS != s) {               // Состояние пина изменилось?
  prevS = s;                      // Если изменилось, запомнили текущее состояние
  if (!flag) {                    // Опрос флага (первое срабатывание?)
  flag = 1;                       // Если первое срабатывание, устанавливаем флаг в 1
  us[count] = now - start;       	// время до первого импульса
  durus[count] = us[count];      	// заносим время до первого импульса в durus[count]
  duration = durus[count];       	// duration = время до фронта первого импульса
  prevUs = now;     	      	  	// заносим текущее машинное время в prevUs
  count++;                        // + 1 в счётчик циклов
  } else {                        // Если не первое срабатывание

                  // Обрабатываем следующие импульсы
  us[count] = now - prevUs;     	// длительность импульса
  durus[count] = (durus[count - 1] + us[count]);      // durus[count] = текущее время всех импульсов
  duration = durus[count - 1];    // duration = время до фронта последнего импульса
  prevUs = now;                   // Запомнили машинное время
  count++;                        // + 1 в счётчик циклов 
  }}                              // Если состояние пина не изменилось, идём дальше
  }                               // Если буфер переполнен, ждём до окончания цикла 60 мс  
 
  if (now - start > 60000) {      // Если время измерения истекло

  			       // Устанавливаем начальные значения  
  if (flag == 0) {                // если импульсов не было
  duration = 0;                   // duration = время до фронта последнего импульса
  for (int i = 0; i < LOG_SIZE; i++) {  // обнуляем массив
  durus[i] = 0; }                 // обнуляем массив
  count = 1;                      // установка счётчика Для вывода на экран
  }                               // если импульсы были, сразу идём сюда
  ledcWrite(ledChannel0, dutyCycle1);    // Останавливаем канал (0 на выходе)
  logOut();                       // Переходим на вывод лога
  }                               // Если время измерения не истекло
  }                               // если не пришла посылка от WeMosArd
  }                               // завершение void loop()
 
  		            // Посылка строба ШИМ
  void setimp() {
  ledc_set_fade_with_step(LEDC_LOW_SPEED_MODE, LEDC_CHANNEL_2, target_duty, scale, cycle_num );
  ledc_set_duty(LEDC_LOW_SPEED_MODE, LEDC_CHANNEL_2, dutyCycle); 
  ledc_update_duty(LEDC_LOW_SPEED_MODE, LEDC_CHANNEL_2);
  ledc_fade_start(LEDC_LOW_SPEED_MODE, LEDC_CHANNEL_2, LEDC_FADE_WAIT_DONE);
  ledc_stop(LEDC_LOW_SPEED_MODE, LEDC_CHANNEL_2, idle_level);
  }

  void initial() {     // Установка начальных значений 

  if (bitRead(reg_IN,1) == 0) {                  // Если А
  IN_PIN = echoPinA;                             // Назначаем IN_PIN А
  } else                                         // если нет, идём дальше
  if (bitRead(reg_IN,2) == 0) {                  // Если B
  IN_PIN = echoPinB;                             // Назначаем IN_PIN B 
  } else                                         // если нет, идём дальше
  if (bitRead(reg_IN,3) == 0) {                  // Если C
  IN_PIN = echoPinC;                             // Назначаем IN_PIN C 
  } else                                         // если нет, идём дальше
  if (bitRead(reg_IN,4) == 0) {                  // Если D
  IN_PIN = echoPinD; }                           // Назначаем IN_PIN D 
  count = 0;                                     // count = 0
  duration = 0;                                  // duration = время до фронта последнего импульса = 0
  for (int i = 0; i < LOG_SIZE; i++) {
  durus[i] = 0;
  }
  flag = 0;                                      // flag = 0
  first = prevS = digitalRead(IN_PIN);           // Запомнили начальное состояние текущего пина
  send_CKL();                                    // Выводим сигнал запуска излучателя
  }

  void logOut() {        // передача полученных результатов на WeMosArd
                         // Читаем напряжение
  analog1 = analogRead(analog1Pin);              // читаем напряжение источника питания
  voltage1 = (analog1*20.9) / 4095;              // пересчитываем число в вольты 

  printSendingData();           // печать отправляемых данных
				                    // Записываем переменные для отправки
  ISPReadings.reg1 = regkn2;
  ISPReadings.reg2 = regm1;
  ISPReadings.reg3 = regm2;
  ISPReadings.reg4 = reg_IN;
  ISPReadings.reg5 = count;
   ISPReadings.voltbat = voltage1;
  ISPReadings.dursig = duration;
  for (int i = 0; i < LOG_SIZE; i++) {
  ISPReadings.arrey1[i] = durus[i];
  }
                                // Отправляем сообщение
  esp_now_send(broadcastAddress,(uint8_t*)&ISPReadings, sizeof(ISPReadings));
  flagrec = 0;                                   // сброс флага посылки 
  }

  void send_CKL() {           // выдача сигналов управления на исполнительное устройство
  			// Установка коэффициента усиления и DC
  digitalWrite(A0R_Pin, bitRead(regout, 2));	   // запись состояния A0 в A0R_Pin
  digitalWrite(A1R_Pin, bitRead(regout, 1));	   // запись состояния A1 в A1R_Pin
  digitalWrite(DC_Pin, !bitRead(regout, 4));	   // запись состояния DC в DC_Pin

  			// Выдача частоты и строба
  if (bitRead(regout,5) == 0) {                  // Если A0, то коэф. = 1
  freqCLK = 10000;                               // Задаём частоту для 8 импульсов
  } else                                         // Если 1, идём мимо
  if (bitRead(regout,3) == 0 ) {                 // Если 0, то 16 импульсов
  freqCLK = 5000;                                // Задаём частоту для 16 импульсов
  } else {                                       // Если 1, идём мимо
  freqCLK = 20000;                               // Задаём частоту для 4 импульсов
  }
  ledcWrite(ledChannel0, dutyCycle0);            // Запускаем частоту 80 кГц
  ledc_set_freq(LEDC_LOW_SPEED_MODE, LEDC_TIMER_1, freqCLK); // Устанавливаем частоту
  setimp();                                      // Запускаем строб
  start = micros();
  }

  /*            7       6       5       4       3       2       1       0        

     regm1     X36      X6      X1   12/24V     -    16 имп   8 имп   4 имп                          
    
     reg_IN     -       -       -       D       C       B       A       -      

     regkn2   ЭКРАН     -      АВТО     D       C       B       A     СРЕДА 

       -        -       -       -      L1       L2      L3     L4       -        

       -        -       -       -     КАН 4   КАН 3    КАН 2   КАН 1    -          

     regout     -       -      A0T     DC      A1T     A0R     A1R      -                           
        0x3A                    1       1       1       0       1       0 

                               A0T   DC_Pin    A1T   A0R_Pin  A1R_Pin   0 
    активный уровень            0       1       0       0       0       0

  */                                        



 
